linkgrammar_get_version
linkgrammar_get_configuration
linkgrammar_open_data_file
linkgrammar_get_dict_version
linkgrammar_get_dict_locale
dictionary_create_lang
dictionary_create_default_lang
dictionary_get_lang
dictionary_delete
dictionary_get_data_dir
dictionary_set_data_dir
boolean_dictionary_lookup
dictionary_word_is_known
dictionary_lookup_list
free_lookup_list
dict_display_word_expr
dict_display_word_info
print_dictionary_data
parse_options_create
parse_options_delete
parse_options_set_verbosity
parse_options_get_verbosity
parse_options_set_debug
parse_options_get_debug
parse_options_set_test
parse_options_get_test
parse_options_set_linkage_limit
parse_options_get_linkage_limit
parse_options_set_disjunct_cost
parse_options_get_disjunct_cost
parse_options_set_min_null_count
parse_options_get_min_null_count
parse_options_set_max_null_count
parse_options_get_max_null_count
parse_options_set_islands_ok
parse_options_get_islands_ok
parse_options_set_spell_guess
parse_options_get_spell_guess
parse_options_set_short_length
parse_options_get_short_length
parse_options_set_max_memory
parse_options_get_max_memory
parse_options_set_max_parse_time
parse_options_get_max_parse_time
parse_options_set_cost_model_type
parse_options_get_cost_model_type
parse_options_set_perform_pp_prune
parse_options_get_perform_pp_prune
parse_options_set_use_sat_parser
parse_options_get_use_sat_parser
parse_options_timer_expired
parse_options_print_total_time
parse_options_memory_exhausted
parse_options_resources_exhausted
parse_options_set_all_short_connectors
parse_options_get_all_short_connectors
parse_options_set_repeatable_rand
parse_options_get_repeatable_rand
parse_options_reset_resources
parse_options_set_display_morphology
parse_options_get_display_morphology
parse_options_set_dialect
parse_options_get_dialect
sentence_create
sentence_delete
sentence_split
sentence_parse
sentence_length
sentence_null_count
sentence_num_linkages_found
sentence_num_valid_linkages
sentence_num_linkages_post_processed
sentence_num_violations
sentence_disjunct_cost
sentence_link_cost
sentence_display_wordgraph
linkage_create
linkage_delete
linkage_get_num_words
linkage_get_num_links
linkage_get_link_lword
linkage_get_link_rword
linkage_get_link_length
linkage_get_link_label
linkage_get_link_llabel
linkage_get_link_rlabel
linkage_get_link_num_domains
linkage_get_link_domain_names
linkage_get_disjunct_str
linkage_get_disjunct_cost
linkage_get_words
linkage_get_word
linkage_get_word_byte_start
linkage_get_word_byte_end
linkage_get_word_char_start
linkage_get_word_char_end
linkage_print_constituent_tree
linkage_free_constituent_tree_str
linkage_print_diagram
linkage_free_diagram
linkage_print_disjuncts
linkage_free_disjuncts
linkage_print_links_and_domains
linkage_free_links_and_domains
linkage_print_postscript
linkage_free_postscript
linkage_print_pp_msgs
linkage_free_pp_msgs
linkage_unused_word_cost
linkage_disjunct_cost
linkage_link_cost
linkage_get_violation_name
lg_error_set_handler
lg_error_set_handler_data
lg_error_formatmsg
lg_error_printall
lg_error_clearall
lg_error_flush
lg_exp_stringify
prt_error
regex_tokenizer_test
utf8_strwidth
